import sys
from twisted.internet import protocol, defer
from zope.interface.verify import verifyObject
from twistedhl7.receiver import IHL7Receiver


class MinimalLowerLayerProtocol(protocol.Protocol):
    """
    Minimal Lower-Layer Protocol (MLLP) takes the form:

        <VT>[HL7 Message]<FS><CR>

    References:

    .. [1] http://www.hl7standards.com/blog/2007/05/02/hl7-mlp-minimum-layer-protocol-defined/
    .. [2] http://www.hl7standards.com/blog/2007/02/01/ack-message-original-mode-acknowledgement/
    """

    _buffer = ''
    start_block = '\x0b'  # <VT>, vertical tab
    end_block = '\x1c'  # <FS>, file separator
    carriage_return = '\x0d'  # <CR>, \r

    def dataReceived(self, data):

        # success callback
        def onSuccess(message):
            self.writeMessage(message)

        # try to find a complete message(s) in the combined the buffer and data
        messages = (self._buffer + data).split(self.end_block)
        # whatever is in the last chunk is an uncompleted message, so put back
        # into the buffer
        self._buffer = messages.pop(-1)

        for raw_message in messages:
            # strip the rest of the MLLP shell from the HL7 message
            raw_message = raw_message.strip(self.start_block + self.carriage_return)

            # only pass messages with data
            if len(raw_message) > 0:
                # convert into unicode, parseMessage expects decoded string
                raw_message = self.factory.decode(raw_message)
                message_container = self.factory.parseMessage(raw_message)

                # error callback (defined here, since error depends on
                # current message).  rejects the message
                def onError(err):
                    reject = message_container.err(err)
                    self.writeMessage(reject)
                    return err

                # have the factory create a deferred and pass the message
                # to the approriate IHL7Receiver instance
                d = self.factory.handleMessage(message_container)
                d.addCallback(onSuccess)
                d.addErrback(onError)

    def writeMessage(self, message):
        if message is None:
            return
        # convert back to a byte string
        message = self.factory.encode(message)
        # wrap message in payload container
        self.transport.write(
            self.start_block + message + self.end_block + self.carriage_return
        )


class MLLPFactory(protocol.ServerFactory):
    protocol = MinimalLowerLayerProtocol

    def __init__(self, receiver):
        verifyObject(IHL7Receiver, receiver)
        self.receiver = receiver
        encoding = receiver.getCodec()
        if isinstance(encoding, tuple):
            encoding, encoding_errors = encoding
        else:
            encoding_errors = None
        self.encoding = encoding or sys.getdefaultencoding()
        self.encoding_errors = encoding_errors or 'strict'

    def parseMessage(self, message_str):
        return self.receiver.parseMessage(message_str)

    def handleMessage(self, message_container):
        # IHL7Receiver allows implementations to return a Deferred or the
        # result, so ensure we return a Deferred here
        return defer.maybeDeferred(self.receiver.handleMessage, message_container)

    def decode(self, value):
        # turn value into unicode using the receiver's declared codec
        if isinstance(value, str):
            return value.decode(self.encoding, self.encoding_errors)
        return unicode(value)

    def encode(self, value):
        # turn value into byte string using the receiver's declared codec
        if isinstance(value, unicode):
            return value.encode(self.encoding, self.encoding_errors)
        return value



from zope.interface import Interface, implements
from twisted.python import log
from twisted.internet import defer

import hl7


class MessageContainer(object):
    """Base class for messages returned from :py:meth:`twistedhl7.receiver.IHL7Receiver.parseMessage`
    and passed to :py:meth:`twistedhl7.receiver.IHL7Receiver.handleMessage`
    """
    def __init__(self, raw_message):
        """Initialize a message with ``raw_message`` - an unparsed HL7 message
        (the MLLP wrapping around the HL7 message will be removed).
        The message will be in unicode, using the codec from
        :py:meth:`twistedhl7.receiver.IHL7Receiver.getCodec` to decode the message.
        """
        self.raw_message = raw_message

    def ack(self, ack_code='AA'):
        """Return unicode acknowledgement message, or None for no ACK.

        ``ack_code`` options are one of `AA` (accept), `AR` (reject), `AE` (error)

        :rtype: unicode
        """
        return None

    def err(self, failure):
        """Handle a twisted errback :py:class:`twisted.python.failure.Failure` ``failure``.
        Subclasses can override to log errors or handle them in a different way.
        Default implementation returns a rejection ACK.

        :rtype: unicode
        """
        # reject the message
        return self.ack(ack_code='AR')


class HL7MessageContainer(MessageContainer):
    """Message implementation that parses using `python-hl7 <http://python-hl7.readthedocs.org>`_"""
    def __init__(self, raw_message):
        super(HL7MessageContainer, self).__init__(raw_message)
        self.message = hl7.parse(raw_message)

    def ack(self, ack_code='AA'):
        """Return HL7 ACK created from the source message.

        :rtype: unicode
        """
        return unicode(self.message.create_ack(ack_code))


class IHL7Receiver(Interface):
    """Interface that must be implemented by MLLP protocol receiver instances"""

    def parseMessage(raw_message):
        """Clients should parse the message and return an instance of :py:class:`twistedhl7.receiver.MessageContainer` or subclass.

        :rtype: :py:class:`twistedhl7.receiver.MessageContainer`
        """
        pass

    def handleMessage(message_container):
        """Clients must implement ``handleMessage``, which takes a ``message_container``
        argument that is the :py:class:`twistedhl7.receiver.MessageContainer` instance
        returned from :py:meth:`twistedhl7.receiver.IHL7Receiver.parseMessage`.
        The implementation, if non-blocking, may directly return the ack/nack
        message or can return the ack/nack within a
        :py:class:`twisted.internet.defer.Deferred`. If the implementation
        involves any blocking code, the implementation must return the result as
        :py:class:`twisted.internet.defer.Deferred` (possibly by using
        :py:func:`twisted.internet.threads.deferToThread`), to prevent the event
        loop from being blocked.
        """
        pass

    def getCodec():
        """Clients should return the codec name [1]_ and error handling scheme [2]_,
        used when decoding into unicode.

        :rtype: tuple(codec, errors)

        .. [1] http://docs.python.org/library/codecs.html#standard-encodings
        .. [2] https://docs.python.org/2/library/codecs.html#codec-base-classes
        """
        pass


class AbstractReceiver(object):
    """Abstract base class implementation of :py:class:`twistedhl7.receiver.IHL7Receiver`"""
    implements(IHL7Receiver)

    message_cls = MessageContainer

    def parseMessage(self, raw_message):
        return self.message_cls(raw_message)

    def getCodec(self):
        return None, None


class AbstractHL7Receiver(AbstractReceiver):
    """Abstract base class implementation of :py:class:`twistedhl7.receiver.IHL7Receiver`

    :rtype: :py:class:`twistedhl7.receiver.HL7MessageContainer`
    """
    message_cls = HL7MessageContainer


class LoggingReceiver(AbstractHL7Receiver):
    """Simple MLLP receiver implementation that logs and ACKs messages."""
    def handleMessage(self, message_container):
        log.msg(message_container.raw_message)
        return defer.succeed(message_container.ack())



from setuptools import setup

setup(
    name='twisted-hl7',
    version='0.2.2',
    author='John Paulett',
    author_email='john@paulett.org',
    description='Async Twisted server implementation of Health Level 7 MLLP',
    url='http://twisted-hl7.readthedocs.org',
    license='BSD',
    platforms=['POSIX', 'Windows'],
    keywords=['HL7', 'Health Level 7', 'healthcare', 'health care',
              'medical record', 'twisted'],
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Communications',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking'
    ],
    packages=['twistedhl7', 'twisted.plugins'],
    package_data={
        "twisted": ["plugins/mllp_plugin.py"]
    },
    install_requires=[
        # require twisted, but allow client to require specific version
        'twisted',
        'hl7>=0.3.1',
    ],
)



from zope.interface import implements
from zope.interface.verify import verifyClass

from twisted.python import usage, reflect
from twisted.plugin import IPlugin
from twisted.internet import endpoints
from twisted.application.service import IServiceMaker
from twisted.application import internet

DEFAULT_ENDPOINT = "tcp:2575"
DEFAULT_RECEIVER = "twistedhl7.receiver.LoggingReceiver"


class Options(usage.Options):
    """Define the options accepted by the ``twistd mllp`` plugin"""
    synopsis = "[mllp options]"

    optParameters = [
        ['endpoint', 'e', DEFAULT_ENDPOINT, 'The string endpoint on which to listen.'],
        ['receiver', 'r', DEFAULT_RECEIVER, 'A twistedhl7.receiver.IHL7Receiver subclass to handle messages.'],
    ]

    longdesc = """\
Starts an MLLP server. If no arguments are specified,
it will be a demo server that logs and ACKs each message received."""


class MLLPServiceMaker(object):
    """Service maker for the MLLP server."""
    implements(IServiceMaker, IPlugin)
    tapname = "mllp"
    description = "HL7 MLLP server."
    options = Options

    def makeService(self, options):
        """Construct a server using MLLPFactory.

        :rtype: :py:class:`twisted.application.internet.StreamServerEndpointService`
        """
        from twisted.internet import reactor
        from twistedhl7.mllp import IHL7Receiver, MLLPFactory

        receiver_name = options['receiver']
        receiver_class = reflect.namedClass(receiver_name)
        verifyClass(IHL7Receiver, receiver_class)
        factory = MLLPFactory(receiver_class())
        endpoint = endpoints.serverFromString(reactor, options['endpoint'])
        server = internet.StreamServerEndpointService(endpoint, factory)
        server.setName(u"mllp-{0}".format(receiver_name))
        return server

serviceMaker = MLLPServiceMaker()
